@@ -1,5 +1,5 @@
 use crossbeam_channel::unbounded;
-use profiler_symbol_server::start_server;
+use profiler_symbol_server::{start_server, PortSelection};
 use serde_json::to_writer;
 use std::fs::File;
 use std::path::{Path, PathBuf};
@@ -8,7 +8,9 @@ use std::thread;
 use std::time::{Duration, Instant};
 use structopt::StructOpt;
 
+#[allow(deref_nullptr)]
 mod dyld_bindings;
+
 mod gecko_profile;
 mod proc_maps;
 mod process_launcher;
@@ -28,30 +30,43 @@ use task_profiler::TaskProfiler;
 #[derive(Debug, StructOpt)]
 #[structopt(
     name = "perfrecord",
-    about = r#"Run a command and record a CPU profile of its execution.
+    about = r#"
+perfrecord is a sampling CPU profiler.
+Run a command, record a CPU profile of its execution, and open the profiler UI.
 
 EXAMPLES:
-    perfrecord ./yourcommand args
-    perfrecord --launch-when-done ./yourcommand args
-    perfrecord -o prof.json ./yourcommand args
-    perfrecord --launch prof.json"#
+    # Default usage:
+    perfrecord ./yourcommand yourargs
+
+    # Alternative usage: Save profile to file for later viewing, and then load it.
+    perfrecord --save-only -o prof.json ./yourcommand yourargs
+    perfrecord --load prof.json
+"#
 )]
 struct Opt {
-    /// Launch the profiler after recording and display the collected profile.
-    #[structopt(long = "launch-when-done")]
-    launch_when_done: bool,
+    /// Do not open the profiler UI.
+    #[structopt(short, long = "no-open")]
+    no_open: bool,
 
-    /// Sampling interval, in seconds
-    #[structopt(short = "i", long = "interval", default_value = "0.001")]
-    interval: f64,
+    /// Do not run a local server after recording.
+    #[structopt(short, long = "save-only")]
+    save_only: bool,
+
+    /// Sampling frequency, in Hz
+    #[structopt(short, long, default_value = "1000")]
+    frequency: f64,
 
     /// Limit the recorded time to the specified number of seconds
     #[structopt(short = "t", long = "time-limit")]
     time_limit: Option<f64>,
 
     /// The port to use for the local web server
-    #[structopt(short, long, default_value = "3000")]
-    port: u16,
+    #[structopt(short, long, default_value = "3000+")]
+    port: String,
+
+    /// Print debugging output.
+    #[structopt(short, long)]
+    verbose: bool,
 
     /// Save the collected profile to this file.
     #[structopt(
@@ -62,17 +77,13 @@ struct Opt {
     )]
     output_file: PathBuf,
 
-    /// If neither --launch nor --serve are specified, profile this command.
+    /// Profile the execution of this command. Ignored if --load is specified.
     #[structopt(subcommand)]
     rest: Option<Subcommands>,
 
-    /// Don't record. Instead, launch the profiler with the selected file in your default browser.
-    #[structopt(short = "l", long = "launch", parse(from_os_str))]
-    file_to_launch: Option<PathBuf>,
-
-    /// Don't record. Instead, serve the selected file from a local webserver.
-    #[structopt(short = "s", long = "serve", parse(from_os_str))]
-    file_to_serve: Option<PathBuf>,
+    /// Don't record. Instead, load the specified file in the profiler.
+    #[structopt(short = "l", long = "load", parse(from_os_str))]
+    load: Option<PathBuf>,
 }
 
 #[derive(Debug, PartialEq, StructOpt)]
@@ -83,44 +94,79 @@ enum Subcommands {
 
 fn main() -> Result<(), MachError> {
     let opt = Opt::from_args();
-    let open_in_browser = opt.file_to_launch.is_some();
-    let file_for_launching_or_serving = opt.file_to_launch.or(opt.file_to_serve);
-    if let Some(file) = file_for_launching_or_serving {
-        start_server_main(&file, opt.port, open_in_browser);
+    let open_in_browser = !opt.no_open;
+    let port_selection = match PortSelection::try_from_str(&opt.port) {
+        Ok(p) => p,
+        Err(e) => {
+            eprintln!(
+                "Could not parse port as <u16> or <u16>+, got port {}, error: {}",
+                opt.port, e
+            );
+            std::process::exit(1)
+        }
+    };
+    let server_props = ServerProps {
+        port_selection,
+        verbose: opt.verbose,
+        open_in_browser,
+    };
+    if let Some(file) = opt.load {
+        start_server_main(&file, server_props);
         return Ok(());
     }
     if let Some(Subcommands::Command(command)) = opt.rest {
         if !command.is_empty() {
             let time_limit = opt.time_limit.map(Duration::from_secs_f64);
-            let interval = Duration::from_secs_f64(opt.interval);
+            if opt.frequency <= 0.0 {
+                eprintln!(
+                    "Error: sampling frequency must be greater than zero, got {}",
+                    opt.frequency
+                );
+                std::process::exit(1);
+            }
+            let interval = Duration::from_secs_f64(1.0 / opt.frequency);
             let exit_status = start_recording(
                 &opt.output_file,
-                opt.port,
                 &command,
                 time_limit,
                 interval,
-                opt.launch_when_done,
+                !opt.save_only,
+                server_props,
             )?;
             std::process::exit(exit_status.code().unwrap_or(0));
         }
     }
-    println!("Error: missing command\n");
+    eprintln!("Error: missing command\n");
     Opt::clap().print_help().unwrap();
+    println!();
     std::process::exit(1);
 }
 
+#[derive(Clone, Debug)]
+struct ServerProps {
+    port_selection: PortSelection,
+    verbose: bool,
+    open_in_browser: bool,
+}
+
 #[tokio::main]
-async fn start_server_main(file: &Path, port: u16, open_in_browser: bool) {
-    start_server(file, port, open_in_browser).await;
+async fn start_server_main(file: &Path, props: ServerProps) {
+    start_server(
+        file,
+        props.port_selection,
+        props.verbose,
+        props.open_in_browser,
+    )
+    .await;
 }
 
 fn start_recording(
     output_file: &Path,
-    port: u16,
     args: &[String],
     time_limit: Option<Duration>,
     interval: Duration,
-    launch_when_done: bool,
+    serve_when_done: bool,
+    server_props: ServerProps,
 ) -> Result<ExitStatus, MachError> {
     let (saver_sender, saver_receiver) = unbounded();
     let output_file = output_file.to_owned();
@@ -130,8 +176,8 @@ fn start_recording(
         to_writer(file, &profile_builder.to_json()).expect("Couldn't write JSON");
 
         // Reuse the saver thread as the server thread.
-        if launch_when_done {
-            start_server_main(&output_file, port, true);
+        if serve_when_done {
+            start_server_main(&output_file, server_props);
         }
     });
 
